#ifndef __reactphysics3d_c_wrapper_h__
#define __reactphysics3d_c_wrapper_h__

#include "reactphysics3d.h"

#include "constraint/ContactPoint.h"
#include "collision/ContactManifold.h"

#ifdef BUILD_DLL
 #define CAPI __declspec(dllexport)
#else
 #define CAPI
#endif


#ifdef __cplusplus
extern "C" {
#endif

typedef struct CollisionCBInfo {
  rp3d::ContactManifoldListElement* contactManifoldElements;
  /// Pointer to the first body of the contact
  rp3d::CollisionBody* body1;
  /// Pointer to the second body of the contact
  rp3d::CollisionBody* body2;
  /// Pointer to the proxy shape of first body
  const rp3d::ProxyShape* proxyShape1;
  /// Pointer to the proxy shape of second body
  const rp3d::ProxyShape* proxyShape2;
} CollisionCBInfo;

typedef rp3d::decimal (RaycastCB)  (const rp3d::RaycastInfo& rcInfo);
typedef void          (OverlapCB)  (rp3d::CollisionBody* body);
typedef void          (CollisionCB)(const CollisionCBInfo & ccbInfo);

#ifdef __cplusplus
}
#endif



class ImpRaycastCallback : public  rp3d::RaycastCallback {
  public:
  void setRaycastCallback(RaycastCB* cb){
    mRaycastCB = cb;
  }
  RaycastCB* getRaycastCallback(void){
    return mRaycastCB;
  }
  virtual  rp3d::decimal notifyRaycastHit(const rp3d::RaycastInfo& info) {
    rp3d::decimal ret=0.0;
    if (mRaycastCB!=NULL) ret = mRaycastCB(info);
    return ret;
  }

  private:
  RaycastCB* mRaycastCB = NULL;
};

class ImpOverlapCallback : public rp3d::OverlapCallback {
  public:
  void setOverlapCallback(OverlapCB* cb){
    mOverlapCB = cb;
  }
  OverlapCB* getOverlapCallback(void){
    return mOverlapCB;
  }
  /// This method will be called for each reported overlapping bodies
  virtual void notifyOverlap(rp3d::CollisionBody* body) {
    if (mOverlapCB!=NULL) mOverlapCB(body);
  }
  private:
  OverlapCB* mOverlapCB = NULL;
};

class ImpCollisionCallback : public rp3d::CollisionCallback {
  public:
  void setCollisionCallback(CollisionCB* cb){
    mCollisionCB = cb;
  }
  CollisionCB* getCollisionCallback(void){
    return mCollisionCB;
  }

  /// This method will be called for each reported contact point
  virtual void notifyContact(const rp3d::CollisionCallback::CollisionCallbackInfo& ccInfo){
    if (mCollisionCB!=NULL) {
      CollisionCBInfo ccbInfo;
      ccbInfo.contactManifoldElements = ccInfo.contactManifoldElements;
      ccbInfo.body1 = ccInfo.body1;
      ccbInfo.body2 = ccInfo.body2;
      ccbInfo.proxyShape1 = ccInfo.proxyShape1;
      ccbInfo.proxyShape2 = ccInfo.proxyShape2;
      mCollisionCB(ccbInfo);
    }
  }
  private:
  CollisionCB* mCollisionCB = NULL;
};

#ifdef __cplusplus
extern "C" {
#endif

/**************
* C callbacks *
**************/
void CAPI SetRaycastCallback(RaycastCB* cb);
RaycastCB* CAPI GetRaycastCallback(void);

void CAPI SetOverlapCallback(OverlapCB* cb);
OverlapCB* CAPI GetOverlapCallback(void);

void CAPI SetCollisionCallback(CollisionCB* cb);
CollisionCB* CAPI GetCollisionCallback(void);


/// Return the next element in the linked-list
rp3d::ContactManifoldListElement* CAPI ContactManifoldListElementGetNext(rp3d::ContactManifoldListElement* manifoldListElement);
/// Return the contact manifold
rp3d::ContactManifold* CAPI ContactManifoldListElementGetContactManifold(rp3d::ContactManifoldListElement* manifoldListElement);


/// Return a pointer to the first proxy shape of the contact
rp3d::ProxyShape* CAPI ContactManifoldGetShape1(rp3d::ContactManifold* cm);
/// Return a pointer to the second proxy shape of the contact
rp3d::ProxyShape* CAPI ContactManifoldGetShape2(rp3d::ContactManifold* cm);
/// Return a pointer to the first body of the contact manifold
rp3d::CollisionBody* CAPI ContactManifoldGetBody1(rp3d::ContactManifold* cm);
/// Return a pointer to the second body of the contact manifold
rp3d::CollisionBody* CAPI ContactManifoldGetBody2(rp3d::ContactManifold* cm);
/// Return the number of contact points in the manifold
rp3d::int8 CAPI ContactManifoldGetNbContactPoints(rp3d::ContactManifold* cm);
/// Return a pointer to the first contact point of the manifold
rp3d::ContactPoint* CAPI ContactManifoldGetContactPoints(rp3d::ContactManifold* cm);
/// Return a pointer to the previous element in the linked-list
rp3d::ContactManifold* CAPI ContactManifoldGetPrevious(rp3d::ContactManifold* cm);
/// Return a pointer to the next element in the linked-list
rp3d::ContactManifold* CAPI ContactManifoldGetNext(rp3d::ContactManifold* cm);

/***************
* ContactPoint *
***************/
/// Return the previous contact point in the linked list
rp3d::ContactPoint* CAPI ContactPointGetPrevious(rp3d::ContactPoint* cp);
/// Return the next contact point in the linked list
rp3d::ContactPoint* CAPI ContactPointGetNext(rp3d::ContactPoint* cp);
/// Return the normal vector of the contact
rp3d::Vector3 CAPI ContactPointGetNormal(rp3d::ContactPoint* cp);
/// Return the contact point on the first proxy shape in the local-space of the proxy shape
const rp3d::Vector3& CAPI ContactPointGetLocalPointOnShape1(rp3d::ContactPoint* cp);
/// Return the contact point on the second proxy shape in the local-space of the proxy shape
const rp3d::Vector3& CAPI ContactPointGetLocalPointOnShape2(rp3d::ContactPoint* cp);
/// Return the cached penetration impulse
rp3d::decimal CAPI ContactPointGetPenetrationImpulse(rp3d::ContactPoint* cp);
/// Return true if the contact is a resting contact
bool CAPI ContactPointGetIsRestingContact(rp3d::ContactPoint* cp);
/// Return the penetration depth
rp3d::decimal CAPI ContactPointGetPenetrationDepth(rp3d::ContactPoint* cp);

/*************
* Quaternion *
*************/
// constructors
/// Return a identity quaternion
rp3d::Quaternion CAPI QuaternionIdentity(void);
/// Return a quaternion constructed from Euler angles (in radians)
rp3d::Quaternion CAPI QuaternionFromEulerAnglesXYZ(rp3d::decimal rx,
                                             rp3d::decimal ry,
                                             rp3d::decimal rz);
rp3d::Quaternion CAPI QuaternionFromEulerAngles(const rp3d::Vector3& rot);
rp3d::Quaternion CAPI QuaternionFromMatrix(const rp3d::Matrix3x3 m);

// methods
void CAPI QuaternionInitWithEulerAnglesXYZ(rp3d::Quaternion & quat,
                                           rp3d::decimal rx,
                                           rp3d::decimal ry,
                                           rp3d::decimal rz);
void CAPI QuaternionInitWithEulerAngles(rp3d::Quaternion & quat,
                                        const rp3d::Vector3& rot);

void CAPI QuaternionGetRotationAngleAxisXYZ(const rp3d::Quaternion & quat,
                                            rp3d::decimal &angle,
                                            rp3d::decimal &ax,
                                            rp3d::decimal &ay,
                                            rp3d::decimal &az);
void CAPI QuaternionGetRotationAngleAxis(const rp3d::Quaternion & quat,
                                         rp3d::decimal &angle,
                                         rp3d::Vector3 &axis);
void CAPI QuaternionMulWith(rp3d::Quaternion& a, const rp3d::Quaternion& b);

/// Set the quaternion to zero
void CAPI QuaternionSetToZero(rp3d::Quaternion& q);
/// Set to the identity quaternion
void CAPI QuaternionSetToIdentity(rp3d::Quaternion& q);
/// Return the length of the quaternion
rp3d::decimal CAPI QuaternionGetLength(rp3d::Quaternion& q);
/// Return the square of the length of the quaternion
rp3d::decimal CAPI QuaternionGetLengthSquare(rp3d::Quaternion& q);
/// Normalize the quaternion
void CAPI QuaternionNormalize(rp3d::Quaternion& q);
rp3d::Quaternion CAPI QuaternionNormalized(rp3d::Quaternion& q);
/// Inverse the quaternion
void CAPI QuaternionInverse(rp3d::Quaternion& q);
/// Return the unit quaternion
rp3d::Quaternion CAPI QuaternionGetUnit(const rp3d::Quaternion& q);
/// Return the conjugate quaternion
rp3d::Quaternion CAPI QuaternionGetConjugate(const rp3d::Quaternion& q);
/// Return the inverse of the quaternion
rp3d::Quaternion CAPI QuaternionGetInverse(const rp3d::Quaternion& q);
/// Return the orientation matrix corresponding to this quaternion
rp3d::Matrix3x3 CAPI QuaternionGetMatrix(const rp3d::Quaternion& q);
/// Dot product between two quaternions
rp3d::decimal CAPI QuaternionGetDot(const rp3d::Quaternion& a, const rp3d::Quaternion& b);
/// Compute the spherical linear interpolation between two quaternions
rp3d::Quaternion CAPI QuaternionSlerp(const rp3d::Quaternion& a, const rp3d::Quaternion& b, rp3d::decimal t);
/// Overloaded operator for addition with assignment
void CAPI QuaternionAddToThis(rp3d::Quaternion& q, const rp3d::Quaternion& a);
/// Overloaded operator for substraction with assignment
void CAPI QuaternionSubFromThis(rp3d::Quaternion& q, const rp3d::Quaternion& a);
/// Overloaded operator for the multiplication with a constant
void CAPI QuaternionMulThis(rp3d::Quaternion& q, rp3d::decimal n);
/// Overloaded operator for the addition
rp3d::Quaternion CAPI QuaternionAdd(const rp3d::Quaternion& a, const rp3d::Quaternion& b);
/// Overloaded operator for the substraction
rp3d::Quaternion CAPI QuaternionSub(const rp3d::Quaternion& a, const rp3d::Quaternion& b);
/// Overloaded operator for the multiplication
rp3d::Quaternion CAPI QuaternionMul(const rp3d::Quaternion& a, const rp3d::Quaternion& b);
/// Overloaded operator for the multiplication with a vector
rp3d::Vector3 CAPI QuaternionMulVector(const rp3d::Quaternion& q, const rp3d::Vector3& point);

/************
* Transform *
************/
rp3d::Transform CAPI TransformFromFromPosMatrix(const rp3d::Vector3& position,
                                                const rp3d::Matrix3x3& orientation);

rp3d::Transform CAPI TransformFromPosRotXYZ(rp3d::decimal px,
                                            rp3d::decimal py,
                                            rp3d::decimal pz,
                                            rp3d::decimal rx,
                                            rp3d::decimal ry,
                                            rp3d::decimal rz);
rp3d::Transform CAPI TransformFromPosRot(const rp3d::Vector3& pos,
                                         const rp3d::Vector3& rot);

const rp3d::Vector3 CAPI & TransformGetPosition(const rp3d::Transform& t);
void CAPI TransformGetPositionXYZ(const rp3d::Transform& t, rp3d::decimal & px, rp3d::decimal & py, rp3d::decimal & pz);
void CAPI TransformSetPosition(rp3d::Transform& t, const rp3d::Vector3& position);
void CAPI TransformSetPositionXYZ(rp3d::Transform& t, rp3d::decimal px, rp3d::decimal py, rp3d::decimal pz);

const rp3d::Quaternion CAPI & TransformGetOrientation(const rp3d::Transform& t);
void CAPI TransformSetOrientation(rp3d::Transform& t, const rp3d::Quaternion& o);
void CAPI TransformSetOrientationFromEulerXYZ(rp3d::Transform& t,
                                         rp3d::decimal rx,
                                         rp3d::decimal ry,
                                         rp3d::decimal rz);

void CAPI TransformSetOpenGLMatrix(rp3d::Transform& t, rp3d::decimal* glm);
void CAPI TransformGetOpenGLMatrix(rp3d::Transform& t, rp3d::decimal* glm);

/// Set the transform to the identity transform
void CAPI TransformSetToIdentity(rp3d::Transform& t);
/// Return the inverse of the transform
rp3d::Transform CAPI TransformGetInverse(const rp3d::Transform& t);
/// Return an interpolated transform
rp3d::Transform CAPI TransformInterpolate(const rp3d::Transform& a, const rp3d::Transform& b, rp3d::decimal t);
/// Return the identity transform
rp3d::Transform CAPI TransformGetIdentity(rp3d::Transform& t);
/// Return the transformed vector
rp3d::Vector3 CAPI TransformMulVector(const rp3d::Transform& t, const rp3d::Vector3& v);
/// Operator of multiplication of a transform with another one
rp3d::Transform CAPI TransformMul(const rp3d::Transform& a, const rp3d::Transform& b);

/************
* Matrix3x3 *
************/

/// constructor
rp3d::Matrix3x3 CAPI CreateMatrix(rp3d::decimal a1, rp3d::decimal a2, rp3d::decimal a3,
                                  rp3d::decimal b1, rp3d::decimal b2, rp3d::decimal b3,
                                  rp3d::decimal c1, rp3d::decimal c2, rp3d::decimal c3);
/// Return the 3x3 identity matrix
rp3d::Matrix3x3 CAPI CreateIdentityMatrix();

/// methods
/// Set the matrix to the identity matrix
void CAPI MatrixSetToIdentity(rp3d::Matrix3x3 m);
/// Set the matrix to zero
void CAPI MatrixSetToZero(rp3d::Matrix3x3 m);
/// Return a column
rp3d::Vector3 CAPI MatrixGetColumn(const rp3d::Matrix3x3 m, int i);
/// Return a row
rp3d::Vector3 CAPI MatrixGetRow(const rp3d::Matrix3x3 m,int i);
/// Return the transpose matrix
rp3d::Matrix3x3 CAPI MatrixGetTranspose(const rp3d::Matrix3x3 m);
/// Return the determinant of the matrix
rp3d::decimal CAPI MatrixGetDeterminant(const rp3d::Matrix3x3 m);
/// Return the trace of the matrix
rp3d::decimal CAPI MatrixGetTrace(const rp3d::Matrix3x3 m);
/// Return the inverse matrix
rp3d::Matrix3x3 CAPI MatrixGetInverse(const rp3d::Matrix3x3 m);
/// Return the matrix with absolute values
rp3d::Matrix3x3 CAPI MatrixGetAbsoluteMatrix(const rp3d::Matrix3x3 m);
/// Overloaded operator for addition
rp3d::Matrix3x3 CAPI MatrixAdd(const rp3d::Matrix3x3& a, const rp3d::Matrix3x3& b);
/// Overloaded operator for substraction
rp3d::Matrix3x3 CAPI MatrixSub(const rp3d::Matrix3x3& a, const rp3d::Matrix3x3& b);
/// Overloaded operator for the negative of the matrix
rp3d::Matrix3x3 CAPI MatrixNeg(const rp3d::Matrix3x3& m);
/// Overloaded operator for multiplication with a matrix
void CAPI MatrixMulThis(rp3d::Matrix3x3& m, rp3d::decimal nb);
/// Overloaded operator for matrix multiplication
rp3d::Matrix3x3 CAPI MatrixMul(const rp3d::Matrix3x3& a, const rp3d::Matrix3x3& b);
/// Overloaded operator for multiplication with a vector
rp3d::Vector3 CAPI MatrixMulVector(const rp3d::Matrix3x3& m, const rp3d::Vector3& v);

/*******
* AABB *
********/
/// Constructor
rp3d::AABB CAPI CreateAABB(const rp3d::Vector3& vmin, const rp3d::Vector3& vmax);
/// Create and return an AABB for a triangle
rp3d::AABB CAPI CreateAABBForTriangle(const rp3d::Vector3* trianglePoints);

/// methods
/// Return the center point
rp3d::Vector3 CAPI AABBGetCenter(const rp3d::AABB& a);
/// Return the minimum coordinates of the AABB
const rp3d::Vector3& CAPI AABBGetMin(const rp3d::AABB& a);
/// Set the minimum coordinates of the AABB
void CAPI AABBSetMin(rp3d::AABB& a, const rp3d::Vector3& vmin);
/// Return the maximum coordinates of the AABB
const rp3d::Vector3& CAPI AABBGetMax(const rp3d::AABB& a);
/// Set the maximum coordinates of the AABB
void CAPI AABBSetMax(rp3d::AABB& a, const rp3d::Vector3& vmax);
/// Return the size of the AABB in the three dimension x, y and z
rp3d::Vector3 CAPI AABBGetExtent(const rp3d::AABB& a);
/// Inflate each side of the AABB by a given size
void CAPI AABBInflate(rp3d::AABB& a, rp3d::decimal dx, rp3d::decimal dy, rp3d::decimal dz);
/// Return true if the current AABB is overlapping with the AABB in argument
bool CAPI AABBTestCollision(const rp3d::AABB& a, const rp3d::AABB& b);
/// Return the volume of the AABB
rp3d::decimal CAPI AABBGetVolume(const rp3d::AABB& a);
/// Merge the AABB in parameter with the current one
void CAPI AABBMergeWithAABB(rp3d::AABB& a, const rp3d::AABB& b);
/// Replace the current AABB with a new AABB that is the union of two AABBs in parameters
void CAPI AABBMergeTwoAABBs(rp3d::AABB& a, const rp3d::AABB& b, const rp3d::AABB& c);
/// Return true if the current AABB contains the AABB given in parameter
bool CAPI AABBContains(const rp3d::AABB& a, const rp3d::AABB& b);
/// Return true if a point is inside the AABB
bool CAPI AABBContainsVector(const rp3d::AABB& a, const rp3d::Vector3& v);
/// Return true if the AABB of a triangle intersects the AABB
bool CAPI AABBTestCollisionTriangleAABB(const rp3d::AABB& a, const rp3d::Vector3* trianglePoints);
/// Return true if the ray intersects the AABB
bool CAPI AABBTestRayIntersect(const rp3d::AABB& a, const rp3d::Ray& r);


/******
* Ray *
*******/


/*****************
* CollisionWorld *
*****************/
// constructor
rp3d::CollisionWorld * CAPI CreateCollisionWorld();
void CAPI DestroyCollisionWorld(rp3d::CollisionWorld * &p);

// CollisionBody constructors
rp3d::CollisionBody* CAPI CollisionWorldCreateCollisionBodyFromPosRotXYZ(
  rp3d::CollisionWorld *cw,
  rp3d::decimal px, rp3d::decimal py, rp3d::decimal pz,
  rp3d::decimal rx, rp3d::decimal ry, rp3d::decimal rz);
rp3d::CollisionBody* CAPI CollisionWorldCreateCollisionBodyFromPosRot(
  rp3d::CollisionWorld *cw,
  const rp3d::Vector3 & pos,
  const rp3d::Vector3 & rot);
rp3d::CollisionBody* CAPI CollisionWorldCreateCollisionBody(
  rp3d::CollisionWorld *cw,
  const rp3d::Transform& t);
void CAPI CollisionWorldDestroyCollisionBody(
  rp3d::CollisionWorld *cw,
  rp3d::CollisionBody* &cb);

// methods
void CAPI CollisionWorldLinecast(rp3d::CollisionWorld *cw,
                           const rp3d::Vector3& startPoint,
                           const rp3d::Vector3& endPoint,
                           rp3d::ushort raycastWithCategoryMaskBits=0xFFFF);
void CAPI CollisionWorldRaycast(rp3d::CollisionWorld *cw,
                                const rp3d::Ray& ray,
                                rp3d::ushort raycastWithCategoryMaskBits=0xFFFF);


/// Test if the AABBs of two bodies overlap
bool CAPI CollisionWorldTestBodyAABBOverlap(rp3d::CollisionWorld *cw,const rp3d::CollisionBody* body1, const rp3d::CollisionBody* body2);
/// Report all the bodies which have an AABB that overlaps with the AABB in parameter
void CAPI CollisionWorldTestAABBOverlap(rp3d::CollisionWorld *cw,const rp3d::AABB& aabb, rp3d::ushort categoryMaskBits=0xFFFF);
/// Return true if two bodies overlap
bool CAPI CollisionWorldTestBodyOverlap(rp3d::CollisionWorld *cw,rp3d::CollisionBody* body1, rp3d::CollisionBody* body2);
/// Report all the bodies that overlap with the body in parameter
void CAPI CollisionWorldTestOverlap(rp3d::CollisionWorld *cw,rp3d::CollisionBody* body, rp3d::ushort categoryMaskBits=0xFFFF);
/// Test and report collisions between two bodies
void CAPI CollisionWorldTestBodyCollision(rp3d::CollisionWorld *cw,rp3d::CollisionBody* body1, rp3d::CollisionBody* body2);
/// Test and report collisions between a body and all the others bodies of the world
void CAPI CollisionWorldTestCollision(rp3d::CollisionWorld *cw,rp3d::CollisionBody* body, rp3d::ushort categoryMaskBits=0xFFFF);
/// Test and report collisions between all shapes of the world
void CAPI CollisionWorldTestAllCollision(rp3d::CollisionWorld *cw);

rp3d::AABB CAPI CollisionWorldGetWorldAABB(rp3d::CollisionWorld *cw, const rp3d::ProxyShape* proxyShape);

/***************************************
* DynamicsWorld extends CollisionWorld *
****************************************/
// constructors
rp3d::DynamicsWorld * CAPI CreateDynamicsWorldXYZ(rp3d::decimal gx,
                                                  rp3d::decimal gy,
                                                  rp3d::decimal gz);
rp3d::DynamicsWorld * CAPI CreateDynamicsWorld(const rp3d::Vector3 & g);
void CAPI DestroyDynamicsWorld(rp3d::DynamicsWorld* &dw);

#if 0
// CollisionBody constructors
rp3d::CollisionBody* CAPI DynamicWorldCreateCollisionBodyFromPosRotXYZ(
  rp3d::DynamicsWorld *dw,
  rp3d::decimal px, rp3d::decimal py, rp3d::decimal pz,
  rp3d::decimal rx, rp3d::decimal ry, rp3d::decimal rz);
rp3d::CollisionBody* CAPI DynamicWorldCreateCollisionBodyFromPosRot(
  rp3d::DynamicsWorld *dw,
  const rp3d::Vector3 & pos,
  const rp3d::Vector3 & rot);
rp3d::CollisionBody* CAPI DynamicWorldCreateCollisionBody(
  rp3d::DynamicsWorld *dw,
  const rp3d::Transform& t);
void CAPI DynamicWorldDestroyCollisionBody(
  rp3d::DynamicsWorld *dw,
  rp3d::CollisionBody* &cb);
#endif

// methods
void CAPI DynamicsWorldUpdate(rp3d::DynamicsWorld *dw,rp3d::decimal timeStep);

rp3d::uint CAPI DynamicsWorldGetNbIterationsVelocitySolver(rp3d::DynamicsWorld *dw);
void CAPI DynamicsWorldSetNbIterationsVelocitySolver(rp3d::DynamicsWorld *dw,rp3d::uint nbIterations);

rp3d::uint CAPI DynamicsWorldGetNbIterationsPositionSolver(rp3d::DynamicsWorld *dw);
void CAPI DynamicsWorldSetNbIterationsPositionSolver(rp3d::DynamicsWorld *dw,rp3d::uint nbIterations);

rp3d::Vector3 CAPI DynamicsWorldGetGravity(rp3d::DynamicsWorld *dw);
void CAPI DynamicsWorldSetGravity(rp3d::DynamicsWorld *dw,rp3d::Vector3& gravity);

bool CAPI DynamicsWorldIsGravityEnabled(rp3d::DynamicsWorld *dw);
void CAPI DynamicsWorldSetIsGratityEnabled(rp3d::DynamicsWorld *dw,bool isGravityEnabled);

rp3d::uint CAPI DynamicsWorldGetNbRigidBodies(rp3d::DynamicsWorld *dw);

bool CAPI DynamicsWorldIsSleepingEnabled(rp3d::DynamicsWorld *dw);
void CAPI DynamicsWorldEnableSleeping(rp3d::DynamicsWorld *dw,bool isSleepingEnabled);

rp3d::decimal CAPI DynamicsWorldGetSleepLinearVelocity(rp3d::DynamicsWorld *dw);
void CAPI DynamicsWorldSetSleepLinearVelocity(rp3d::DynamicsWorld *dw,rp3d::decimal sleepLinearVelocity);

rp3d::decimal CAPI DynamicsWorldGetSleepAngularVelocity(rp3d::DynamicsWorld *dw);
void CAPI DynamicsWorldSetSleepAngularVelocity(rp3d::DynamicsWorld *dw,rp3d::decimal sleepAngularVelocity);

rp3d::decimal CAPI DynamicsWorldGetTimeBeforeSleep(rp3d::DynamicsWorld *dw);
void CAPI DynamicsWorldSetTimeBeforeSleep(rp3d::DynamicsWorld *dw,rp3d::decimal timeBeforeSleep);

/// Create a joint between two bodies in the world and return a pointer to the new joint
rp3d::Joint* CAPI DynamicsWorldCreateJoint(rp3d::DynamicsWorld *dw,const rp3d::JointInfo& jointInfo);
rp3d::FixedJoint* CAPI DynamicsWorldCreateFixedJoint(rp3d::DynamicsWorld *dw,const rp3d::JointInfo& jointInfo);
rp3d::BallAndSocketJoint* CAPI DynamicsWorldCreateBallAndSocketJoint(rp3d::DynamicsWorld *dw,const rp3d::JointInfo& jointInfo);
rp3d::HingeJoint* CAPI DynamicsWorldCreateHingeJoint(rp3d::DynamicsWorld *dw,const rp3d::JointInfo& jointInfo);
rp3d::SliderJoint* CAPI DynamicsWorldCreateSliderJoint(rp3d::DynamicsWorld *dw,const rp3d::JointInfo& jointInfo);

void CAPI DynamicsWorldDestroyJoint(rp3d::DynamicsWorld *dw,rp3d::Joint* & joint);

rp3d::uint CAPI DynamicsWorldGetNbJoints(rp3d::DynamicsWorld *dw);



/*******
* Body *
*******/
/// Return the ID of the body
rp3d::bodyindex CAPI BodyGetId(rp3d::Body * b);
/// Return whether or not the body is allowed to sleep
bool CAPI BodyIsAllowedToSleep(rp3d::Body * b);
/// Set whether or not the body is allowed to go to sleep
void CAPI BodySetIsAllowedToSleep(rp3d::Body * b, bool isAllowedToSleep);
/// Set the variable to know whether or not the body is sleeping
void CAPI BodySetIsSleeping(rp3d::Body * b, bool isSleeping);
/// Return whether or not the body is sleeping
bool CAPI BodyIsSleeping(rp3d::Body * b);
/// Return true if the body is active
bool CAPI BodyIsActive(rp3d::Body * b);
/// Set whether or not the body is active
void CAPI BodySetIsActive(rp3d::Body * b, bool isActive);
/// Return a pointer to the user data attached to this body
void* CAPI BodyGetUserData(rp3d::Body * b);
/// Attach user data to this body
void CAPI BodySetUserData(rp3d::Body * b, void* userData);

/*****************************
* CollisionBody extends Body *
******************************/

void CAPI CollisionBodySetPosRotXYZ(rp3d::CollisionBody* cb,
                                    rp3d::decimal px,
                                    rp3d::decimal py,
                                    rp3d::decimal pz,
                                    rp3d::decimal rx,
                                    rp3d::decimal ry,
                                    rp3d::decimal rz);
void CAPI CollisionBodySetPosRot(rp3d::CollisionBody* cb,
                                 const rp3d::Vector3 & pos,
                                 const rp3d::Vector3 & rot);

const rp3d::Transform CAPI & CollisionBodyGetTransform(rp3d::CollisionBody* cb);
void CAPI CollisionBodySetTransform(rp3d::CollisionBody* cb, const rp3d::Transform& t);

rp3d::BodyType CAPI CollisionBodyGetType(rp3d::CollisionBody* cb);
void CAPI CollisionBodySetType(rp3d::CollisionBody* cb, rp3d::BodyType bt);

void CAPI CollisionBodySetIsActive(rp3d::CollisionBody* cb,bool isActive);
bool CAPI CollisionBodyTestPointInside(rp3d::CollisionBody* cb,const rp3d::Vector3& worldPoint);
bool CAPI CollisionBodyRaycast(rp3d::CollisionBody* cb,const rp3d::Ray& ray, rp3d::RaycastInfo& raycastInfo);
bool CAPI CollisionBodyTestAABBOverlap(rp3d::CollisionBody* cb,const rp3d::AABB& worldAABB);
rp3d::AABB CAPI CollisionBodyGetAABB(rp3d::CollisionBody* cb);
rp3d::ProxyShape* CAPI CollisionBodyGetProxyShapesList(rp3d::CollisionBody* cb);
const rp3d::ProxyShape* CAPI CollisionBodyGetProxyShapesListC(rp3d::CollisionBody* cb);
/// Return the world-space coordinates of a point given the local-space coordinates of the body
rp3d::Vector3 CAPI CollisionBodyGetWorldPoint(rp3d::CollisionBody* cb,const rp3d::Vector3& localPoint);
/// Return the world-space vector of a vector given in local-space coordinates of the body
rp3d::Vector3 CAPI CollisionBodyGetWorldVector(rp3d::CollisionBody* cb,const rp3d::Vector3& localVector);
/// Return the body local-space coordinates of a point given in the world-space coordinates
rp3d::Vector3 CAPI CollisionBodyGetLocalPoint(rp3d::CollisionBody* cb,const rp3d::Vector3& worldPoint);
/// Return the body local-space coordinates of a vector given in the world-space coordinates
rp3d::Vector3 CAPI CollisionBodyGetLocalVector(rp3d::CollisionBody* cb,const rp3d::Vector3& worldVector);


/// ProxyShape constructors
rp3d::ProxyShape* CAPI CollisionBodyAddCollisionShapePosRotXYZ(
  rp3d::CollisionBody* cb,
  rp3d::CollisionShape* cs,
  rp3d::decimal px, rp3d::decimal py, rp3d::decimal pz,
  rp3d::decimal rx, rp3d::decimal ry, rp3d::decimal rz);

rp3d::ProxyShape* CAPI CollisionBodyAddCollisionShapePosRot(
  rp3d::CollisionBody* cb,
  rp3d::CollisionShape* cs,
  const rp3d::Vector3 & pos,
  const rp3d::Vector3 & rot);

rp3d::ProxyShape* CAPI CollisionBodyAddCollisionShape(
  rp3d::CollisionBody* cb,
  rp3d::CollisionShape* cs,
  const rp3d::Transform& t);

rp3d::ProxyShape* CAPI CollisionBodyAddCollisionShapeAtCenter(rp3d::CollisionBody* cb,
                                                              rp3d::CollisionShape* cs);

void CAPI CollisionBodyRemoveCollisionShape(rp3d::CollisionBody* cb, const rp3d::ProxyShape* proxyShape);



/**********************************
* RigidBody extends CollisionBody *
***********************************/
// constructors
rp3d::RigidBody* CAPI DynamicsWorldCreateRigidBodyPosRotXYZ(
  rp3d::DynamicsWorld *dw,
  rp3d::decimal px, rp3d::decimal py, rp3d::decimal pz,
  rp3d::decimal rx, rp3d::decimal ry, rp3d::decimal rz);

rp3d::RigidBody* CAPI DynamicsWorldCreateRigidBodyPosRot(
  rp3d::DynamicsWorld *dw,
  const rp3d::Vector3 & pos,
  const rp3d::Vector3 & rot);

rp3d::RigidBody* CAPI DynamicsWorldCreateRigidBody(
  rp3d::DynamicsWorld *dw,
  const rp3d::Transform& t);

void CAPI DynamicsWorldDestroyRigidBody(
  rp3d::DynamicsWorld *dw,
  rp3d::RigidBody* &rb);

// methods
void CAPI RigidBodySetType(rp3d::RigidBody* rb,rp3d::BodyType bt);
rp3d::BodyType CAPI RigidBodyGetType(const rp3d::RigidBody* rb);

/// Set the current position and orientation
void CAPI RigidBodySetTransform(rp3d::RigidBody* rb, const rp3d::Transform& t);
void CAPI RigidBodySetTransformPosRotXYZ(
  rp3d::RigidBody *rb,
  rp3d::decimal px, rp3d::decimal py, rp3d::decimal pz,
  rp3d::decimal rx, rp3d::decimal ry, rp3d::decimal rz);
void CAPI RigidBodySetTransformPosRot(
  rp3d::RigidBody *rb,
  const rp3d::Vector3 & pos,
  const rp3d::Vector3 & rot);



/// Return the mass of the body
rp3d::decimal CAPI RigidBodyGetMass(rp3d::RigidBody* rb);
/// Set the mass of the rigid body
void CAPI RigidBodySetMass(rp3d::RigidBody* rb, rp3d::decimal mass);

/// Return the linear velocity
rp3d::Vector3 CAPI RigidBodyGetLinearVelocity(rp3d::RigidBody* rb);
void CAPI RigidBodyGetLinearVelocityXYZ(rp3d::RigidBody* rb,
                                        rp3d::decimal & vx,
                                        rp3d::decimal & vy,
                                        rp3d::decimal & vz);
/// Set the linear velocity of the body.
void CAPI RigidBodySetLinearVelocity(rp3d::RigidBody* rb, const rp3d::Vector3& linearVelocity);
void CAPI RigidBodySetLinearVelocityXYZ(rp3d::RigidBody* rb,
                                        rp3d::decimal vx,
                                        rp3d::decimal vy,
                                        rp3d::decimal vz);
/// Return the angular velocity
rp3d::Vector3 CAPI RigidBodyGetAngularVelocity(rp3d::RigidBody* rb);
void CAPI RigidBodyGetAngularVelocityXYZ(rp3d::RigidBody* rb,
                                        rp3d::decimal & vx,
                                        rp3d::decimal & vy,
                                        rp3d::decimal & vz);
/// Set the angular velocity.
void CAPI RigidBodySetAngularVelocity(rp3d::RigidBody* rb, const rp3d::Vector3& angularVelocity);
void CAPI RigidBodySetAngularVelocityXYZ(rp3d::RigidBody* rb,
                                         rp3d::decimal vx,
                                         rp3d::decimal vy,
                                         rp3d::decimal vz);


/// Set the variable to know whether or not the body is sleeping
void CAPI RigidBodySetIsSleeping(rp3d::RigidBody* rb,
                                 bool isSleeping);
/// Set the local inertia tensor of the body (in body coordinates)
void CAPI RigidBodySetInertiaTensorLocal(rp3d::RigidBody* rb,
                                         const rp3d::Matrix3x3& inertiaTensorLocal);
/// Set the inverse local inertia tensor of the body (in body coordinates)
void CAPI RigidBodySetInverseInertiaTensorLocal(rp3d::RigidBody* rb,
                                                const rp3d::Matrix3x3& inverseInertiaTensorLocal);
/// Get the inverse local inertia tensor of the body (in body coordinates)
const rp3d::Matrix3x3& CAPI RigidBodyGetInverseInertiaTensorLocal(rp3d::RigidBody* rb);
/// Return the inverse of the inertia tensor in world coordinates.
rp3d::Matrix3x3 CAPI RigidBodyGetInertiaTensorInverseWorld(rp3d::RigidBody* rb);
/// Set the local center of mass of the body (in local-space coordinates)
void CAPI RigidBodySetCenterOfMassLocal(rp3d::RigidBody* rb,
                                        const rp3d::Vector3& centerOfMassLocal);
void CAPI RigidBodySetCenterOfMassLocalXYZ(rp3d::RigidBody* rb,
                                           rp3d::decimal x,
                                           rp3d::decimal y,
                                           rp3d::decimal z);
/// Return true if the gravity needs to be applied to this rigid body
bool CAPI RigidBodyIsGravityEnabled(rp3d::RigidBody* rb);
/// Set the variable to know if the gravity is applied to this rigid body
void CAPI RigidBodyEnableGravity(rp3d::RigidBody* rb,
                                 bool isEnabled);
/// Return a reference to the material properties of the rigid body
rp3d::Material& CAPI RigidBodyGetMaterial(rp3d::RigidBody* rb);
/// Set a new material for this rigid body
void CAPI RigidBodySetMaterial(rp3d::RigidBody* rb,
                               const rp3d::Material& material);
/// Return the linear velocity damping factor
rp3d::decimal CAPI RigidBodyGetLinearDamping(rp3d::RigidBody* rb);
/// Set the linear damping factor
void CAPI RigidBodySetLinearDamping(rp3d::RigidBody* rb,
                                    rp3d::decimal linearDamping);
/// Return the angular velocity damping factor
rp3d::decimal CAPI RigidBodyGetAngularDamping(rp3d::RigidBody* rb);
/// Set the angular damping factor
void CAPI RigidBodySetAngularDamping(rp3d::RigidBody* rb,
                                     rp3d::decimal angularDamping);
/// Apply an external force to the body at its center of mass.
void CAPI RigidBodyApplyForceToCenterOfMass(rp3d::RigidBody* rb,
                                            const rp3d::Vector3& force);
void CAPI RigidBodyApplyForceToCenterOfMassXYZ(rp3d::RigidBody* rb,
                                               rp3d::decimal x, rp3d::decimal y, rp3d::decimal z);

/// Apply an external force to the body at a given point (in world-space coordinates).
void CAPI RigidBodyApplyForce(rp3d::RigidBody* rb,
                              const rp3d::Vector3& force,
                              const rp3d::Vector3& point);
void CAPI RigidBodyApplyForceXYZ(rp3d::RigidBody* rb,
                                 rp3d::decimal fx, rp3d::decimal fy, rp3d::decimal fz,
                                 rp3d::decimal px, rp3d::decimal py, rp3d::decimal pz);

/// Apply an external torque to the body.
void CAPI RigidBodyApplyTorque(rp3d::RigidBody* rb,const rp3d::Vector3& torque);
void CAPI RigidBodyApplyTorqueXYZ(rp3d::RigidBody* rb, rp3d::decimal x, rp3d::decimal y,rp3d::decimal z);

/// Return the first element of the linked list of joints involving this body
const rp3d::JointListElement* CAPI RigidBodyGetJointsListC(rp3d::RigidBody* rb);
/// Return the first element of the linked list of joints involving this body
rp3d::JointListElement* CAPI RigidBodyGetJointsList(rp3d::RigidBody* rb);

rp3d::ProxyShape* CAPI RigidBodyAddCollisionShapePosRotXYZ(
  rp3d::RigidBody* rb,
  rp3d::CollisionShape* cs,
  rp3d::decimal px, rp3d::decimal py, rp3d::decimal pz,
  rp3d::decimal rx, rp3d::decimal ry, rp3d::decimal rz,
  rp3d::decimal mass);

rp3d::ProxyShape* CAPI RigidBodyAddCollisionShapePosRot(
  rp3d::RigidBody* rb,
  rp3d::CollisionShape* cs,
  const rp3d::Vector3 & pos,
  const rp3d::Vector3 & rot,
  rp3d::decimal mass);


rp3d::ProxyShape* CAPI RigidBodyAddCollisionShape(
  rp3d::RigidBody* rb,
  rp3d::CollisionShape* cs,
  const rp3d::Transform& t,
  rp3d::decimal mass);

rp3d::ProxyShape* CAPI RigidBodyAddCollisionShapeAtCenter(rp3d::RigidBody* rb,
                                                          rp3d::CollisionShape* cs,
                                                          rp3d::decimal mass);

void CAPI RigidBodyRemoveCollisionShape(rp3d::RigidBody* rb,
                                        const rp3d::ProxyShape* proxyShape);
/********
* Joint *
********/
/// Return the reference to the body 1
rp3d::RigidBody* CAPI JointGetBody1(rp3d::Joint* j);
/// Return the reference to the body 2
rp3d::RigidBody* CAPI JointGetBody2(rp3d::Joint* j);
/// Return true if the constraint is active
bool CAPI JointIsActive(rp3d::Joint* j);
/// Return the type of the constraint
rp3d::JointType CAPI JointGetType(rp3d::Joint* j);
/// Return true if the collision between the two bodies of the joint is enabled
bool CAPI JointIsCollisionEnabled(rp3d::Joint* j);
/// Return the id of the joint
rp3d::uint CAPI JointGetId(rp3d::Joint* j);

/// Return a pointer to the user data attached to this joint
void* CAPI JointGetUserData(rp3d::Joint * j);
/// Attach user data to this body
void CAPI JointSetUserData(rp3d::Joint * j, void* userData);

/*************
* HingeJoint *
*************/
/// Return true if the limits or the joint are enabled
bool CAPI HingeJointIsLimitEnabled(rp3d::HingeJoint * hj);
/// Enable/Disable the limits of the joint
void CAPI HingeJointEnableLimit(rp3d::HingeJoint * hj, bool isLimitEnabled);
/// Return true if the motor of the joint is enabled
bool CAPI HingeJointIsMotorEnabled(rp3d::HingeJoint * hj);
/// Enable/Disable the motor of the joint
void CAPI HingeJointEnableMotor(rp3d::HingeJoint * hj, bool isMotorEnabled);
/// Return the minimum angle limit
rp3d::decimal CAPI HingeJointGetMinAngleLimit(rp3d::HingeJoint * hj);
/// Set the minimum angle limit
void CAPI HingeJointSetMinAngleLimit(rp3d::HingeJoint * hj, rp3d::decimal lowerLimit);
/// Return the maximum angle limit
rp3d::decimal CAPI HingeJointGetMaxAngleLimit(rp3d::HingeJoint * hj);
/// Set the maximum angle limit
void CAPI HingeJointSetMaxAngleLimit(rp3d::HingeJoint * hj, rp3d::decimal upperLimit);
/// Return the motor speed
rp3d::decimal CAPI HingeJointGetMotorSpeed(rp3d::HingeJoint * hj);
/// Set the motor speed
void CAPI HingeJointSetMotorSpeed(rp3d::HingeJoint * hj, rp3d::decimal motorSpeed);
/// Return the maximum motor torque
rp3d::decimal HingeJointGetMaxMotorTorque(rp3d::HingeJoint * hj);
/// Set the maximum motor torque
void CAPI HingeJointSetMaxMotorTorque(rp3d::HingeJoint * hj, rp3d::decimal maxMotorTorque);
/// Return the intensity of the current torque applied for the joint motor
rp3d::decimal CAPI HingeJointGetMotorTorque(rp3d::HingeJoint * hj, rp3d::decimal timeStep);

/**************
* SliderJoint *
**************/
/// Return true if the limits or the joint are enabled
bool CAPI SliderJointIsLimitEnabled(rp3d::SliderJoint* sj);
/// Enable/Disable the limits of the joint
void CAPI SliderJointEnableLimit(rp3d::SliderJoint* sj, bool isLimitEnabled);
/// Return true if the motor of the joint is enabled
bool CAPI SliderJointIsMotorEnabled(rp3d::SliderJoint* sj);
/// Enable/Disable the motor of the joint
void CAPI SliderJointEnableMotor(rp3d::SliderJoint* sj,bool isMotorEnabled);
/// Return the current translation value of the joint
rp3d::decimal CAPI SliderJointGetTranslation(rp3d::SliderJoint* sj);
/// Return the minimum translation limit
rp3d::decimal CAPI SliderJointGetMinTranslationLimit(rp3d::SliderJoint* sj);
/// Set the minimum translation limit
void CAPI SliderJointSetMinTranslationLimit(rp3d::SliderJoint* sj, rp3d::decimal lowerLimit);
/// Return the maximum translation limit
rp3d::decimal CAPI SliderJointGetMaxTranslationLimit(rp3d::SliderJoint* sj);
/// Set the maximum translation limit
void CAPI SliderJointSetMaxTranslationLimit(rp3d::SliderJoint* sj, rp3d::decimal upperLimit);
/// Return the motor speed
rp3d::decimal CAPI SliderJointGetMotorSpeed(rp3d::SliderJoint* sj);
/// Set the motor speed
void CAPI SliderJointSetMotorSpeed(rp3d::SliderJoint* sj, rp3d::decimal motorSpeed);
/// Return the maximum motor force
rp3d::decimal CAPI SliderJointGetMaxMotorForce(rp3d::SliderJoint* sj);
/// Set the maximum motor force
void CAPI SliderJointSetMaxMotorForce(rp3d::SliderJoint* sj,rp3d::decimal maxMotorForce);
/// Return the intensity of the current force applied for the joint motor
rp3d::decimal CAPI SliderJointGetMotorForce(rp3d::SliderJoint* sj, rp3d::decimal timeStep);




/*************
* ProxyShape *
*************/
/// Return the collision shape
const rp3d::CollisionShape* CAPI ProxyShapeGetCollisionShape(rp3d::ProxyShape* ps);
/// Return the parent body
rp3d::CollisionBody* CAPI ProxyShapeGetBody(rp3d::ProxyShape* ps);
/// Return the mass of the collision shape
rp3d::decimal CAPI ProxyShapeGetMass(rp3d::ProxyShape* ps);
/// Return a pointer to the user data attached to this body
void* CAPI ProxyShapeGetUserData(rp3d::ProxyShape* ps);
/// Attach user data to this body
void CAPI ProxyShapeSetUserData(rp3d::ProxyShape* ps,void* userData);
/// Return the local to parent body transform
const rp3d::Transform& CAPI ProxyShapeGetLocalToBodyTransform(rp3d::ProxyShape* ps);
/// Set the local to parent body transform
void CAPI ProxyShapeSetLocalToBodyTransform(rp3d::ProxyShape* ps,const rp3d::Transform& t);
/// Return the local to world transform
const rp3d::Transform CAPI ProxyShapeGetLocalToWorldTransform(rp3d::ProxyShape* ps);
/// Return the AABB of the proxy shape in world-space
const rp3d::AABB CAPI ProxyShapeGetWorldAABB(rp3d::ProxyShape* ps);
/// Test if the proxy shape overlaps with a given AABB
bool CAPI ProxyShapeTestAABBOverlap(rp3d::ProxyShape* ps,const rp3d::AABB& worldAABB);
/// Return true if a point is inside the collision shape
bool CAPI ProxyShapeTestPointInside(rp3d::ProxyShape* ps,const rp3d::Vector3& worldPoint);
/// Raycast method with feedback information
bool CAPI ProxyShapeLinecast(rp3d::ProxyShape* ps,
                             const rp3d::Vector3& startPoint,
                             const rp3d::Vector3& endPoint,
                             rp3d::RaycastInfo& raycastInfo);
bool CAPI ProxyShapeRaycast(rp3d::ProxyShape* ps, const rp3d::Ray& ray, rp3d::RaycastInfo& raycastInfo);
/// Return the collision bits mask
rp3d::ushort CAPI ProxyShapeGetCollideWithMaskBits(rp3d::ProxyShape* ps);
/// Set the collision bits mask
void CAPI ProxyShapeSetCollideWithMaskBits(rp3d::ProxyShape* ps,rp3d::ushort collideWithMaskBits);
/// Return the collision category bits
rp3d::ushort CAPI ProxyShapeGetCollisionCategoryBits(rp3d::ProxyShape* ps);
/// Set the collision category bits
void CAPI ProxyShapeSetCollisionCategoryBits(rp3d::ProxyShape* ps,rp3d::ushort collisionCategoryBits);
/// Return the next proxy shape in the linked list of proxy shapes
rp3d::ProxyShape* CAPI ProxyShapeGetNext(rp3d::ProxyShape* ps);
/// Return the next const proxy shape in the linked list of proxy shapes
const rp3d::ProxyShape* CAPI ProxyShapeGetNextC(rp3d::ProxyShape* ps);


/*****************
* CollisionShape *
******************/

/*************************************
* ConvexShape extends CollisionShape *
*************************************/

/**********************************
* SphereShape extends ConvexShape *
**********************************/
rp3d::SphereShape* CAPI CreateSphereShape(rp3d::decimal radius);
void CAPI DestroySphereShape(rp3d::SphereShape* &s);
/***********************************
* CapsuleShape extends ConvexShape *
************************************/
rp3d::CapsuleShape* CAPI CreateCapsuleShape(rp3d::decimal radius, rp3d::decimal height);
void CAPI DestroyCapsuleShape(rp3d::CapsuleShape* &s);


/*****************************************
* BoxShape extends ConvexPolyhedronShape *
*****************************************/
rp3d::BoxShape* CAPI CreateBoxShape(const rp3d::Vector3& halfExtents);
void CAPI DestroyBoxShape(rp3d::BoxShape* &s);

/**************
* PolygonFace *
**************/
rp3d::PolygonVertexArray::PolygonFace* CAPI CreatePolygonFaces(rp3d::uint nbFaces);
/*********************
* PolygonVertexArray *
*********************/
// constructor
rp3d::PolygonVertexArray* CAPI CreatePolygonVertexArray(
  rp3d::uint nbVertices,
  const void* verticesStart,
  int verticesStride,
  const void* indexesStart,
  int indexesStride,
  rp3d::uint nbFaces,
  rp3d::PolygonVertexArray::PolygonFace* facesStart,
  rp3d::PolygonVertexArray::VertexDataType vertexDataType,
  rp3d::PolygonVertexArray::IndexDataType indexDataType);

void CAPI DestroyPolygonVertexArray(rp3d::PolygonVertexArray* & pva);

// methods
/// Return the vertex data type
rp3d::PolygonVertexArray::VertexDataType CAPI PolygonVertexArrayGetVertexDataType(rp3d::PolygonVertexArray* pva);
/// Return the index data type
rp3d::PolygonVertexArray::IndexDataType CAPI PolygonVertexArrayGetIndexDataType(rp3d::PolygonVertexArray* pva);
/// Return the number of vertices
rp3d::uint CAPI PolygonVertexArrayGetNbVertices(rp3d::PolygonVertexArray* pva);
/// Return the number of faces
rp3d::uint CAPI PolygonVertexArrayGetNbFaces(rp3d::PolygonVertexArray* pva);
/// Return the vertices stride (number of bytes)
int CAPI PolygonVertexArrayGetVerticesStride(rp3d::PolygonVertexArray* pva);
/// Return the indices stride (number of bytes)
int CAPI PolygonVertexArrayGetIndicesStride(rp3d::PolygonVertexArray* pva);
/// Return the vertex index of a given vertex in a face
rp3d::uint CAPI PolygonVertexArrayGetVertexIndexInFace(rp3d::PolygonVertexArray* pva,
                                                  rp3d::uint faceIndex,
                                                  rp3d::uint noVertexInFace);
/// Return a polygon face of the polyhedron
rp3d::PolygonVertexArray::PolygonFace* CAPI PolygonVertexArrayGetPolygonFace(rp3d::PolygonVertexArray* pva,
                                                                        rp3d::uint faceIndex);
/// Return the pointer to the start of the vertices array
const rp3d::uchar* CAPI PolygonVertexArrayGetVerticesStart(rp3d::PolygonVertexArray* pva);
/// Return the pointer to the start of the indices array
const rp3d::uchar* CAPI PolygonVertexArrayGetIndicesStart(rp3d::PolygonVertexArray* pva);

/*****************
* PolyhedronMesh *
*****************/
rp3d::PolyhedronMesh* CAPI CreatePolyhedronMesh(rp3d::PolygonVertexArray* pva);
void CAPI DestroyPolyhedronMesh(rp3d::PolyhedronMesh* & phm);

/************************************************
* ConvexMeshShape extends ConvexPolyhedronShape *
************************************************/
rp3d::ConvexMeshShape * CAPI CreateConvexMeshShape(rp3d::PolyhedronMesh* phm);
rp3d::ConvexMeshShape * CAPI CreateScaledConvexMeshShape(rp3d::PolyhedronMesh* phm,
                                                         rp3d::Vector3 & scale);
rp3d::ConvexMeshShape * CAPI CreateScaledConvexMeshShapeXYZ(rp3d::PolyhedronMesh* phm,
                                                            rp3d::decimal sx,
                                                            rp3d::decimal sy,
                                                            rp3d::decimal sz);
void CAPI DestroyConvexMeshShape(rp3d::ConvexMeshShape* & cms);

/**********************
* TriangleVertexArray *
***********************/
/// Constructor without vertices normals
rp3d::TriangleVertexArray* CAPI CreateTriangleVertexArray(
  rp3d::uint nbVertices,
  const void* verticesStart,
  rp3d::uint verticesStride,
  rp3d::uint nbTriangles,
  const void* indexesStart,
  rp3d::uint indexesStride,
  rp3d::TriangleVertexArray::VertexDataType vertexDataType,
  rp3d::TriangleVertexArray::IndexDataType indexDataType);

/// Constructor with vertices normals
rp3d::TriangleVertexArray* CAPI CreateTriangleVertexNormalArray(
  rp3d::uint nbVertices,
  const void* verticesStart,
  rp3d::uint verticesStride,
  const void* verticesNormalsStart,
  rp3d::uint uverticesNormalsStride,
  rp3d::uint nbTriangles,
  const void* indexesStart,
  rp3d::uint indexesStride,
  rp3d::TriangleVertexArray::VertexDataType vertexDataType,
  rp3d::TriangleVertexArray::NormalDataType normalDataType,
  rp3d::TriangleVertexArray::IndexDataType indexDataType);

void CAPI DestroyTriangleVertexArray(rp3d::TriangleVertexArray* & tva);

// methods
/// Return the vertex data type
rp3d::TriangleVertexArray::VertexDataType CAPI TriangleVertexArrayGetVertexDataType(rp3d::TriangleVertexArray* tva);

/// Return the vertex normal data type
rp3d::TriangleVertexArray::NormalDataType CAPI TriangleVertexArrayGetVertexNormalDataType(rp3d::TriangleVertexArray* tva);

  /// Return the index data type
rp3d::TriangleVertexArray::IndexDataType CAPI TriangleVertexArrayGetIndexDataType(rp3d::TriangleVertexArray* tva);

/// Return the number of vertices
rp3d::uint CAPI TriangleVertexArrayGetNbVertices(rp3d::TriangleVertexArray* tva);
/// Return the number of triangles
rp3d::uint CAPI TriangleVertexArrayGetNbTriangles(rp3d::TriangleVertexArray* tva);
/// Return the vertices stride (number of bytes)
rp3d::uint CAPI TriangleVertexArrayGetVerticesStride(rp3d::TriangleVertexArray* tva);
/// Return the vertex normals stride (number of bytes)
rp3d::uint CAPI TriangleVertexArrayGetVerticesNormalsStride(rp3d::TriangleVertexArray* tva);
/// Return the indices stride (number of bytes)
rp3d::uint CAPI TriangleVertexArrayGetIndicesStride(rp3d::TriangleVertexArray* tva);
/// Return the pointer to the start of the vertices array
const void* CAPI TriangleVertexArrayGetVerticesStart(rp3d::TriangleVertexArray* tva);
/// Return the pointer to the start of the vertex normals array
const void* CAPI TriangleVertexArrayGetVerticesNormalsStart(rp3d::TriangleVertexArray* tva);
/// Return the pointer to the start of the indices array
const void* CAPI TriangleVertexArrayGetIndicesStart(rp3d::TriangleVertexArray* tva);
/// Return the vertices coordinates of a triangle
void CAPI TriangleVertexArrayGetTriangleVertices(rp3d::TriangleVertexArray* tva,
                                                 rp3d::uint triangleIndex,
                                                 rp3d::Vector3* outTriangleVertices);
/// Return the three vertices normals of a triangle
void CAPI TriangleVertexArrayGetTriangleVerticesNormals(rp3d::TriangleVertexArray* tva,
                                                        rp3d::uint triangleIndex,
                                                        rp3d::Vector3* outTriangleVerticesNormals);
/// Return the indices of the three vertices of a given triangle in the array
void CAPI TriangleVertexArrayGetTriangleVerticesIndices(rp3d::TriangleVertexArray* tva,
                                                        rp3d::uint triangleIndex,
                                                        rp3d::uint* outVerticesIndices);
/// Return a vertex of the array
void CAPI TriangleVertexArrayGetVertex(rp3d::TriangleVertexArray* tva,
                                       rp3d::uint vertexIndex,
                                       rp3d::Vector3* outVertex);
/// Return a vertex normal of the array
void CAPI TriangleVertexArrayGetNormal(rp3d::TriangleVertexArray* tva,
                                       rp3d::uint vertexIndex,
                                       rp3d::Vector3* outNormal);

/***************
* TriangleMesh *
***************/
rp3d::TriangleMesh* CAPI CreateTriangleMesh(void);
void CAPI DestroyTriangleMesh(rp3d::TriangleMesh* & tm);

void CAPI TriangleMeshAddSubpart(rp3d::TriangleMesh* tm,
                                 rp3d::TriangleVertexArray* tva);

rp3d::uint CAPI TriangleMeshGetNbSubparts(rp3d::TriangleMesh* tm);
rp3d::TriangleVertexArray* CAPI TriangleMeshGetSubpart(rp3d::TriangleMesh* tm,
                                                       rp3d::uint indexSubpart);

/****************************************
* ConcaveMeshShape extends ConcaveShape *
****************************************/
rp3d::ConcaveMeshShape* CAPI CreateConcaveMeshShape(rp3d::TriangleMesh* tm);
rp3d::ConcaveMeshShape* CAPI CreateScaledConcaveMeshShape(rp3d::TriangleMesh* tm,
                                                          rp3d::Vector3 & scale);
rp3d::ConcaveMeshShape* CAPI CreateScaledConcaveMeshShapeXYZ(rp3d::TriangleMesh* tm,
                                                             rp3d::decimal sx,
                                                             rp3d::decimal sy,
                                                             rp3d::decimal sz);

void CAPI DestroyConcaveMeshShape(rp3d::ConcaveMeshShape* & cms);

/****************************************
* HeightFieldShape extends ConcaveShape *
****************************************/

rp3d::HeightFieldShape* CAPI CreateHeightFieldShape(int nbGridColumns,
                                                    int nbGridRows,
                                                    rp3d::decimal minHeight,
                                                    rp3d::decimal maxHeight,
                                                    const void* heightFieldData,
                                                    rp3d::HeightFieldShape::HeightDataType dataType);

void CAPI DestroyHeightFieldShape(rp3d::HeightFieldShape* & hf);


#ifdef __cplusplus
}
#endif

#endif //! __reactphysics3d_c_wrapper_h__
